/*
 * this handles many different operations used to manipulate a bounding box.
 * while most representations of a bounding box use xy-coordinates for their
 * data, in the vst world the xy part of the data is coordinates, the width
 * and height are absolute. This makes it easy to go from the parent object
 * to a plant object.
 *
 * the primary operation is spliting a box into smaller chunks, either in the
 * vertical or horizontal directions.
 */
#ifndef BBOX_CSDX
#define BBOX_CSDX #1#

/*
opcode bb_toInt, i[][], i[][]
    iBBoxes[][] xin
    iLen = lenarray(iBBoxes)
    iCount = 0
    while iCount < iLen do
        iBB[] = getrow(iBBoxes, iCount)
        iLen2 = lenarray(iBB)
        iCount2 = 0
        while iCount2 < iLen2 do
            iBB[iCount2] = floor(iBB[iCount2] + .5)
            iCount2 += 1
        od
        iCount += 1
    od
    xout iBBoxes
endop
*/

opcode bb_x, i, i[]
	iBB[] xin
	xout iBB[0]
endop

opcode bb_y, i, i[]
	iBB[] xin
	xout iBB[1]
endop

opcode bb_w, i, i[]
	iBB[] xin
	xout iBB[2]
endop

opcode bb_h, i, i[]
	iBB[] xin
	xout iBB[3]
endop

opcode bb_base, i[], ii
    iW, iH xin
    iBBox[] fillarray 0, 0, iW, iH
    xout iBBox
endop

opcode bb_parentContext, i[], i[]
    iBBox[] xin
    iBBox[0] = 0
    iBBox[1] = 0
    xout iBBox
endop

opcode bb_withRotation, i[], i[]
    iBBox[] xin
    iBBox2[] = fillarray(iBBox[0] + iBBox[2], iBBox[1], iBBox[3], iBBox[2])
    xout iBBox2
endop

opcode bb_trim, i[], i[]Si
    iBBox[], SElement, iAmount xin
    iBBox2[] = fillarray(iBBox[0], iBBox[1], iBBox[2], iBBox[3])
    if strcmp(SElement, "x") == 0 then
        ;; because the width is fixed, we have to adjust the width value as well
        iBBox2[0] = iBBox2[0] + iAmount
        iBBox2[2] = iBBox2[2] - iAmount
    elseif strcmp(SElement, "y") == 0 then
        ;; because the height is fixed, we have to adjust the height value as well
        iBBox2[1] = iBBox2[1] + iAmount
        iBBox2[3] = iBBox2[3] - iAmount
    elseif strcmp(SElement, "w") == 0 then
        iBBox2[2] = iBBox2[2] - iAmount
    elseif strcmp(SElement, "h") == 0 then
        iBBox2[3] = iBBox2[3] - iAmount
    else
        prints "bb_trim: unknown element selector '%s'\n", SElement
    endif
    
    xout iBBox2
endop

opcode bb_inset, i[], i[]ii
    iBBox[], iX, iY xin
    iBBox2[] = fillarray(iBBox[0] + iX, iBBox[1] + iY, iBBox[2] - (iX * 2), iBBox[3] - (iY * 2))
    xout iBBox2
endop

opcode bb_expand, i[], i[]ii
    iBBox[], iW, iH xin
    iBBox2[] = fillarray(iBBox[0], iBBox[1], iBBox[2] * iW, iBBox[3] * iH)
    xout iBBox2
endop

/*
opcode bb_change, i[], i[]Sijjj
    iBBox[], SParam, iVal xin
    iBBox2[] init lenarray(iBBox)
    iLen = strlen(SParam)
    if iLen > 1 then
        
    else
        if strcmp(SParam, "x") == 0 then
            ; if moving x, it effects w
            iBBox2[0] = iBBox[0] + iVal
            iBBox2[1] = iBBox[1]
            iBBox2[2] = iBBox[2] - iVal
            iBBox2[3] = iBBox[3]
        elseif strcmp(SParam, "y") == 0 then
            iBBox2[0] = iBBox[0]
            iBBox2[1] = iBBox[1] + iVal
            iBBox2[2] = iBBox[2]
            iBBox2[3] = iBBox[3] - iVal
        elseif strcmp(SParam, "w") == 0 then
            iBBox2[0] = iBBox[0]
            iBBox2[1] = iBBox[1]
            iBBox2[2] = iBBox[2] + iVal
            iBBox2[3] = iBBox[3]
        elseif strcmp(SParam, "h") == 0 then
            iBBox2[0] = iBBox[0]
            iBBox2[1] = iBBox[1]
            iBBox2[2] = iBBox[2]
            iBBox2[3] = iBBox[3] + iVal
        else
            prints "bb_change: unknown option '%s'. no change made.", SParam
            iBBox2[0] = iBBox[0]
            iBBox2[1] = iBBox[1]
            iBBox2[2] = iBBox[2]
            iBBox2[3] = iBBox[3]
        endif
    endif
    xout iBBox2
endop
*/

/*
converts the width and height values into actual pixel locations
 */
opcode bb_getLowerRightCoordinates, ii, i[]
    iBBox[] xin
    iX2 = iBBox[0] + iBBox[2]
    iY2 = iBBox[1] + iBBox[3]
    xout iX2, iY2
endop

/*
converts from coordinates back to width and height. input is NOT
a bounding box, but only a temporary data type used locally.
 */
opcode bb_getWidthHeightFromCoordinates, ii, i[]
    iBBoxX[] xin
    iW = iBBoxX[2] - iBBoxX[0]
    iY = iBBoxX[3] - iBBoxX[1]
    xout iW, iY
endop

/*
returns the differences between two different bounding boxes.
while this could be used as a bounding box, it isn't intended to be one.
 */
opcode bb_getDifference, iiii, i[]i[]
    iBBoxA[], iBBoxB[] xin
    iX = iBBoxA[0] - iBBoxB[0]
    iY = iBBoxA[1] - iBBoxB[1]
    iW = iBBoxA[2] - iBBoxB[2]
    iH = iBBoxA[3] - iBBoxB[3]
    xout iX, iY, iW, iH
endop

/*
this will only effect iBBox[2] and iBBox[3]
scales height and width by a percentage
*/
opcode bb_scaleSizeBy, i[], i[]ij
    iBBox[], iWScale, iHScale xin
    if iHScale == -1 then
        iHScale = iWScale
    endif
    iBBoxOut[] = fillarray(iBBox[0], iBBox[1], iBBox[2] * iWScale, iBBox[3] * iHScale)
    xout iBBoxOut
endop

opcode bb_translateBy, i[], i[]ii
    iBBox[], iXShift, iYShift xin
    iBBoxOut[] = fillarray(iBBox[0] + iXShift, iBBox[1] + iYShift, iBBox[2], iBBox[3])
    xout iBBoxOut
endop

opcode bb_scaleCentered, i[], i[]ij
    iBBox[], iWScale, iHScale xin
    iBBox2[] = bb_scaleSizeBy(iBBox, iWScale, iHScale)
    iBBox2[] = bb_translateBy(iBBox2, (iBBox[2] - iBBox2[2]) / 2, (iBBox[3] - iBBox2[3]) / 2)
    xout iBBox2
endop

opcode bb_pack, i[][], i[]
	iBB[] xin
	iBBoxes[][] init 1, 4
	iBBoxes setrow iBB, 0
	xout iBBoxes
endop

opcode bb_pack, i[][], i[]i[]
	iBB[], iBB2[] xin
	iBBoxes[][] init 2, 4
	iBBoxes setrow iBB, 0
	iBBoxes setrow iBB2, 1
	xout iBBoxes
endop

opcode bb_pack, i[][], i[]i[]i[]
	iBB[], iBB2[], iBB3[] xin
	iBBoxes[][] init 3, 4
	iBBoxes setrow iBB, 0
	iBBoxes setrow iBB2, 1
	iBBoxes setrow iBB3, 2
	xout iBBoxes
endop

opcode bb_pack, i[][], i[]i[]i[]i[]
	iBB[], iBB2[], iBB3[], iBB4[] xin
	iBBoxes[][] init 4, 4
	iBBoxes setrow iBB, 0
	iBBoxes setrow iBB2, 1
	iBBoxes setrow iBB3, 2
	iBBoxes setrow iBB4, 3
	xout iBBoxes
endop

opcode bb_pack, i[][], i[]i[]i[]i[]i[]
	iBB[], iBB2[], iBB3[], iBB4[], iBB5[] xin
	iBBoxes[][] init 5, 4
	iBBoxes setrow iBB, 0
	iBBoxes setrow iBB2, 1
	iBBoxes setrow iBB3, 2
	iBBoxes setrow iBB4, 3
	iBBoxes setrow iBB5, 4
	xout iBBoxes
endop

opcode bb_pack, i[][], i[]i[]i[]i[]i[]i[]
	iBB[], iBB2[], iBB3[], iBB4[], iBB5[], iBB6[] xin
	iBBoxes[][] init 6, 4
	iBBoxes setrow iBB, 0
	iBBoxes setrow iBB2, 1
	iBBoxes setrow iBB3, 2
	iBBoxes setrow iBB4, 3
	iBBoxes setrow iBB5, 4
	iBBoxes setrow iBB6, 5
	xout iBBoxes
endop

/*
this takes a list of bounding boxes and determines the bounding box that encloses them all.
 */
opcode bb_enclosing, i[], i[][]i[]
    iBBoxes[][], iIndices[] xin
    iLen = lenarray(iIndices)
    iBBoxes2[][] init iLen, 4
    iCount = 0
    while iCount < iLen do
        iBBox[] = getrow(iBBoxes, iIndices[iCount])
        iX2, iY2 bb_getLowerRightCoordinates iBBox
        iBCoords[] = fillarray(iBBox[0], iBBox[1], iX2, iY2)
        iBBoxes2 setrow iBCoords, iCount
        
        iCount = iCount + 1
    od
    iXS[] = getcol(iBBoxes2, 0)
    iNewX minarray iXS
    iYS[] = getcol(iBBoxes2, 1)
    iNewY minarray iYS
    iWS[] = getcol(iBBoxes2, 2)
    iNewW maxarray iWS
    iHS[] = getcol(iBBoxes2, 3)
    iNewH maxarray iHS
    iTemp[] = fillarray(iNewX, iNewY, iNewW, iNewH)
    
    iW, iH bb_getWidthHeightFromCoordinates iTemp
    iTemp[2] = iW
    iTemp[3] = iH
    xout iTemp
endop

opcode bb_enclosing, i[], i[][]
    iBBoxes[][] xin
    iLen = lenarray(iBBoxes)
    iIndices[] init iLen
    iCount = 0
    while iCount < iLen do
        iIndices[iCount] = iCount
        iCount += 1
    od
    xout bb_enclosing(iBBoxes, iIndices)
endop

/*
this splits a bounding box into two parts, based on the percentage given.
the percentage represents the size of the first bounding box returned.
 */
opcode bb_split, i[][], i[]Sj
    iBBox[], SOrient, iSplitPoint xin
    iBBox2[][] init 2, 4
    if iSplitPoint < 0 || iSplitPoint > 1 then
        iSplitPoint = .5
    endif
    if strcmp(SOrient, "v") == 0 then
        iXSplit = iBBox[2] * iSplitPoint
        iBBox2 fillarray iBBox[0], iBBox[1], iXSplit, iBBox[3], iBBox[0] + iXSplit, iBBox[1], iBBox[2] - iXSplit, iBBox[3]
    elseif strcmp(SOrient, "h") == 0 then
        iYSplit = iBBox[3] * iSplitPoint
        iBBox2 fillarray iBBox[0], iBBox[1], iBBox[2], iYSplit, iBBox[0], iBBox[1] + iYSplit, iBBox[2], iBBox[3] - iYSplit
    else
        prints "bb_split(a): unknown orientation of '%s'.\n", SOrient
        iXSplit = iBBox[2] * iSplitPoint
        iBBox2 fillarray iBBox[0], iBBox[1], iXSplit, iBBox[3], iBBox[0] + iXSplit, iBBox[1], iBBox[2] - iXSplit, iBBox[3]
    endif
    xout iBBox2
endop

opcode bb_split, i[]i[], i[]Sj
    iBBox[], SOrient, iSplitPoint xin
    iSplitBoxes[][] bb_split iBBox, SOrient, iSplitPoint
    xout getrow(iSplitBoxes, 0), getrow(iSplitBoxes, 1)
endop

/*
splits a bounding box into equal divisions based on the split count.
 */
opcode bb_split, i[][], i[]iS
    iBBox[], iSplits, SOrient xin
    iBounds[][] init iSplits, 4
    if strcmp(SOrient, "v") == 0 then
        iXinc = iBBox[2] / iSplits
        iYinc = 0
        iW = iXinc
        iH = iBBox[3]
    elseif strcmp(SOrient, "h") == 0 then
        iXinc = 0
        iYinc = iBBox[3] / iSplits
        iW = iBBox[2]
        iH = iYinc
    else
        prints "bb_split(b): unknown SOrient option of '%s'\n", SOrient
        iXinc = iBBox[2] / iSplits
        iYinc = 0
        iW = iXinc
        iH = iBBox[3]
    endif
    iCount = 0
    while iCount < iSplits do
        iB[] fillarray iBBox[0] + (iXinc * iCount), iBBox[1] + (iYinc * iCount), iW, iH
        iBounds setrow iB, iCount
        iCount = iCount + 1
    od
    xout iBounds
endop

/*
this is only here to turn this code into single lines...
 */
opcode bb_setvalues, iiiiiii, iiiiiii
    iXSt, iXEn, iXIn, iYSt, iYEn, iYIn, iSwap xin
    
    xout iXSt, iXEn, iXIn, iYSt, iYEn, iYIn, iSwap
endop

/*
splits a bounding box into 'iXSplits X iYSplits' grid of bounding boxes.
*/
opcode bb_split, i[][], i[]iiS
    iBBox[], iXSplits, iYSplits, SOrder xin
    iTotalCount = iXSplits * iYSplits
    iBBoxesOut[][] init iTotalCount, 4
    iCount = 0
    iXStart, iXEnd, iXInc, iYStart, iYEnd, iYInc, iSwap = 0
    if strcmp(SOrder, "lrtb") == 0 then
        iXStart, iXEnd, iXInc, iYStart, iYEnd, iYInc, iSwap bb_setvalues 0, iYSplits, 1, 0, iXSplits, 1, 1
    elseif strcmp(SOrder, "lrbt") == 0 then
        iXStart, iXEnd, iXInc, iYStart, iYEnd, iYInc, iSwap bb_setvalues iYSplits - 1, -1, -1, 0, iXSplits, 1, 1
    elseif strcmp(SOrder, "rltb") == 0 then
        iXStart, iXEnd, iXInc, iYStart, iYEnd, iYInc, iSwap bb_setvalues 0, iYSplits, 1, iXSplits - 1, -1, -1, 1
    elseif strcmp(SOrder, "rlbt") == 0 then
        iXStart, iXEnd, iXInc, iYStart, iYEnd, iYInc, iSwap bb_setvalues iYSplits - 1, -1, -1, iXSplits - 1, -1, -1, 1
    elseif strcmp(SOrder, "tblr") == 0 then
        iXStart, iXEnd, iXInc, iYStart, iYEnd, iYInc, iSwap bb_setvalues 0, iXSplits, 1, 0, iYSplits, 1, 0
    elseif strcmp(SOrder, "tbrl") == 0 then
        iXStart, iXEnd, iXInc, iYStart, iYEnd, iYInc, iSwap bb_setvalues iXSplits - 1, -1, -1, 0, iYSplits, 1, 0
    elseif strcmp(SOrder, "btlr") == 0 then
        iXStart, iXEnd, iXInc, iYStart, iYEnd, iYInc, iSwap bb_setvalues 0, iXSplits, 1, iYSplits - 1, -1, -1, 0
    elseif strcmp(SOrder, "btrl") == 0 then
        iXStart, iXEnd, iXInc, iYStart, iYEnd, iYInc, iSwap bb_setvalues iXSplits - 1, -1, -1, iYSplits - 1, -1, -1, 0
    else
        prints "unknown key '%s'\n", SOrder
        iXStart, iXEnd, iXInc, iYStart, iYEnd, iYInc, iSwap bb_setvalues 0, iYSplits, 1, 0, iXSplits, 1, 1
    endif
    iX = iXStart
    iWidth = iBBox[2] / iXSplits
    iHeight = iBBox[3] / iYSplits
    while iX != iXEnd do
        iY = iYStart
        while iY != iYEnd do
            if iSwap == 0 then
                iXorg = (iX * iWidth) + iBBox[0]
                iYorg = (iY * iHeight) + iBBox[1]
            else
                iYorg = (iX * iHeight) + iBBox[1]
                iXorg = (iY * iWidth) + iBBox[0]
            endif
            iRow[] = fillarray(iXorg, iYorg, iWidth, iHeight)
            iBBoxesOut setrow iRow, iCount
            iCount = iCount + 1
            iY = iY + iYInc
        od
        iX = iX + iXInc
    od
    xout iBBoxesOut
endop

opcode bb_choose, i[][], i[]i[]S
    iBBox[], iSegments[], SAlign xin
    iTotalSegments = 0
    iLen = lenarray(iSegments)
    iOutBoxes[][] init iLen, 4
    iCount = 0
    while iCount < iLen do
        iTotalSegments += iSegments[iCount]
        iCount += 1
    od
    ; prints "bb_choose: segments: %d\n", iTotalSegments
    iFrameBoxes[][] bb_split iBBox, iTotalSegments, SAlign
;    printarray iFrameBoxes
    iCount = 0
    iIndex = 0
    while iCount < iLen do
        iBB[] = getrow(iFrameBoxes, iIndex)
        ; printarray iBB, "%f", "iBB in"
        if strcmp(SAlign, "v") == 0 then
            iBB[] = bb_expand(iBB, iSegments[iCount], 1)
        elseif strcmp(SAlign, "h") == 0 then
            iBB[] = bb_expand(iBB, 1, iSegments[iCount])
        else
            prints "bb_choose: unknown v/h selection (%s). Defaulting to 'v'.\n", SAlign
            iBB[] = bb_expand(iBB, iSegments[iCount], 1)
        endif
        ; printarray iBB, "%f", "iBB out"
        iOutBoxes setrow iBB, iCount
        iIndex += iSegments[iCount]
        iCount += 1
    od
    xout iOutBoxes
endop

/*
opcode bb_splitSegments, i[][], i[]i[]
    iBBox[], iSegList[] xin
    iLen = lenarray(iSegList)
    iBSegments[][] init iLen, 4
    
    iSegments = 0
    iCount = 0
    while iCount < iLen do
        iSegments = iSegments + iSegList[iCount]
        iCount = iCount + 1
    od
    iStep = iBBox[2] / iSegments
    iCount = 0
    iX = 0
    iY = 0
    iW = 0
    iH = iBBox[3]
    while iCount < iLen do
        iW = iSegList[iCount] * iStep
        iBBox2[] = fillarray(iX, iY, iW, iH)
        iBSegments setrow iBBox2, iCount
        iX = iX + iW
        iCount = iCount + 1
    od
    xout iBSegments
endop

opcode bb_splitSegments, i[][], i[]i[]S
    iBBox[], iSegments[], SOrient xin
    iLen = lenarray(iSegments)
    iBSegs[][] init iLen, 4
    iChange = 0
    if strcmp(SOrient, "v") == 0 then
        ; change x
        iChange = 0
    elseif strcmp(SOrient, "h") == 0 then
        ; change y
        iChange = 1
    else
        prints "unknown orientation '%s'\n", SOrient
    endif
    
    iSegs = 0
    iCount = 0
    while iCount < iLen do
        iSegs = iSegs + iSegments[iCount]
        iCount = iCount + 1
    od
    iStep = iBBox[2] / iSegs
    iCount = 0
    iX = 0
    iY = 0
    if iChange == 0 then
        ; change x/w
        iW = iBBox[2]
        iH = 0
    else
        iW = 0
        iH = iBBox[3]
    endif
    while iCount < iLen do
        if iChange == 0 then
            iW = iSegments[iCount] * iStep
        else
            iH = iSegments[iCount] * iStep
        endif
        iBBox2[] = fillarray(iX, iY, iW, iH)
        iBSegs setrow iBBox2, iCount
        if iChange == 0 then
            iX = iX + iW
        else
            iY = iY + iH
        endif
        iCount = iCount + 1
    od
    xout iBSegs
endop
*/

/*
returns the desired bounding box as it relates to the input bounding box.
the desired size does not have to be contained within the anchored box.

the position is determined by a two character string.
the first character is one of "lcr", for left, centre, and right.
the second character is one of "tmb", for top, middle, and bottom.

should this also provide the means of positioning components outside of the
bounding box, but still anchored in someway?

if either the desired sizes is negative, then the value is taken from the
input bounding box.
 */
opcode bb_inside, i[], i[]iiS
    iBBox[], iDesiredXSize, iDesiredYSize, SAlign xin
    if iDesiredXSize <= 0 then
        ; iDesiredXSize = iBBox[2]
        iDesiredXSize = iBBox[2] + iDesiredXSize
    endif
    if iDesiredYSize <= 0 then
        ; iDesiredYSize = iBBox[3]
        iDesiredYSize = iBBox[3] + iDesiredYSize
    endif
    iBBox2[] fillarray 0, 0, iDesiredXSize, iDesiredYSize
    iLeft strchar SAlign, 0
    iTop strchar SAlign, 1
    if iLeft == 108 then ; 'l'
        iBBox2[0] = iBBox[0]
    elseif iLeft == 99 then ; 'c'
        iBBox2[0] = iBBox[0] + ((iBBox[2] / 2) - (iDesiredXSize / 2))
    elseif iLeft == 114 then ; 'r'
        iBBox2[0] = iBBox[0] + iBBox[2] - iDesiredXSize
    else
        iBBox2[0] = iBBox[0]
    endif
    if iTop == 116 then ; 't'
        iBBox2[1] = iBBox[1]
    elseif iTop == 109 then ; 'm'
        iBBox2[1] = iBBox[1] + ((iBBox[3] / 2) - (iDesiredYSize / 2))
    elseif iTop == 98 then ; 'b'
        iBBox2[1] = iBBox[1] + iBBox[3] - iDesiredYSize
    else
        iBBox2[1] = iBBox[1]
    endif
    xout iBBox2
endop

opcode bb_inside, i[], i[]Sij
    iBBox[], SAlign, iXScale, iYScale xin
    if iYScale == -1 then
        iYScale = iXScale
    endif
    iDesiredXSize = iBBox[2] * iXScale
    iDesiredYSize = iBBox[3] * iYScale
    xout bb_inside(iBBox, iDesiredXSize, iDesiredYSize, SAlign)
endop

opcode bb_outside, i[], i[]iiS
    iBBox[], iDesiredXSize, iDesiredYSize, SAlign xin
    ; SHOULD THIS ALSO CHECK FOR NEGATIVE VALUES FOR X AND Y?
    if iDesiredXSize < 0 then
        iDesiredXSize = iBBox[2]
    endif
    if iDesiredYSize < 0 then
        iDesiredYSize = iBBox[3]
    endif
    if strcmp(SAlign, "cm") == 0 then
        ;; illegal option
        prints "illegal option (cm) in bb_alignOutside. defaulting to 'lt'"
        SAlign = "lt"
    endif
    iBBox2[] fillarray 0, 0, iDesiredXSize, iDesiredYSize
    iLeft strchar SAlign, 0
    iTop strchar SAlign, 1
    if iLeft == 108 then ; 'l'
        iBBox2[0] = iBBox[0] - iDesiredXSize
    elseif iLeft == 99 then ; 'c'
        iBBox2[0] = iBBox[0] + ((iBBox[2] / 2) - (iDesiredXSize / 2))
    elseif iLeft == 114 then ; 'r'
        iBBox2[0] = iBBox[0] + iBBox[2]
    else
        iBBox2[0] = iBBox[0]
    endif
    if iTop == 116 then ; 't'
        iBBox2[1] = iBBox[1] - iDesiredYSize
    elseif iTop == 109 then ; 'm'
        iBBox2[1] = iBBox[1] + ((iBBox[3] / 2) - (iDesiredYSize / 2))
    elseif iTop == 98 then ; 'b'
        iBBox2[1] = iBBox[1] + iBBox[3]
    else
        iBBox2[1] = iBBox[1]
    endif
    xout iBBox2
endop

#end ; BBOX_CSDX
